import logging
from argparse import ArgumentParser
from argparse import SUPPRESS

from pyloot import PyLootServer


def parse_args():
    parser = ArgumentParser(
        add_help=False,
        description="Run a local wsgi server for remote in-memory storage of pyloot data",
    )
    parser.add_argument(
        "-h", "--host", default="0.0.0.0", help="Host to listen on. (Default: 0.0.0.0)"
    )
    parser.add_argument(
        "-p",
        "--port",
        default=8000,
        type=int,
        help="Port to listen on. (Default: 8000)",
    )
    parser.add_argument(
        "--help",
        action="help",
        help="show this help message and exit",
        default=SUPPRESS,
    )

    return parser.parse_args()


def main():
    logging.basicConfig(
        format="[%(asctime)s %(name)s %(levelname)s]: %(message)s", level=logging.INFO
    )
    args = parse_args()
    server = PyLootServer()
    server.serve_forever(host=args.host, port=args.port)



import logging
import threading
import time
from typing import Optional
from typing import Set

from pyloot.backends.base import BaseBackend
from pyloot.backends.http import HTTPRemoteBackend
from pyloot.backends.memory import InMemoryBackend
from pyloot.collector import get_object_descriptors
from pyloot.server import PyLootServer
from pyloot.utils import start_thread


logger = logging.getLogger(__name__)


class PyLoot:
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        backend: Optional[BaseBackend] = None,
        interval: int = 30,
    ):
        if backend and (host or port):
            logger.warning("ignoring host and port since backend is present")
        if backend:
            self._backend = backend
        elif host or port:
            self._backend = HTTPRemoteBackend(host=host, port=port)
        else:
            self._backend = InMemoryBackend()

        self._running = False
        self._thread_ended = threading.Event()
        self._server: Optional[PyLootServer] = None
        self._interval = interval

    def start(self):
        """
        Start collecting data in a background thread.

        If gevent is detected, gevent.threadpool.spawn is used.
        Otherwise, threading.Thread is used.
        """
        if self._running:
            logger.warning("pyloot looter thread already started")
            return

        self._running = True

        logger.debug("Starting pyloot looter thread")
        start_thread(self._run)
        logger.debug("Finished starting pyloot looter thread")

    def stop(self, blocking=False):
        self._running = False
        if blocking:
            self._thread_ended.wait()

    def _run(self):
        while self._running:
            try:
                self.collect_objects()
            except Exception:  # pylint: disable=broad-except
                logger.exception("Error collecting objects")
            time.sleep(self._interval)

        self._thread_ended.set()

    def collect_objects(self):
        """
        Collect objects still in `gc.get_objects` after a call to `gc.collect`
        """
        logger.debug("Collecting objects")
        st = time.monotonic()
        ignored = self._backend.get_ids()
        ignore_set: Set[int] = set(ignored)
        del ignored
        ignore_set.add(id(self))
        ignore_set.add(id(ignore_set))
        logger.debug("There are %s ids to ignore", len(ignore_set))
        data = get_object_descriptors(ignore_set)
        dur = time.monotonic() - st
        logger.debug("Collecting objects took %s secs.", dur)
        self._backend.store(data)

    def get_wsgi(self) -> PyLootServer:
        """
        Return a WSGI compatible application serving the PyLoot remote backend and
        and the website.
        :return: ::class::`PyLootServer`
        """
        if not self._server:
            self._server = PyLootServer(backend=self._backend)
        return self._server



from pyloot import collector
from pyloot import InMemoryBackend


def test_store_blank_slate():
    backend = InMemoryBackend()
    test_obj = dict(a=1)
    obj_descr = collector.get_data(test_obj, set())
    backend.store([obj_descr])

    descrs = backend.fetch()
    assert len(descrs) == 1
    assert descrs[0].seen
    old_seen = descrs[0].seen

    history = backend.fetch_history()
    assert len(history) == 1
    assert history[0].counts == [1]

    # duplicate obj should not increase data or count
    backend.store([obj_descr])

    descrs = backend.fetch()
    assert len(descrs) == 1
    assert descrs[0].seen == old_seen

    history = backend.fetch_history()
    assert len(history) == 1
    assert history[0].counts == [1, 1]

    test_obj2 = dict(a=1)
    obj_descr2 = collector.get_data(test_obj2, set())

    # new obj should increase data and count
    backend.store([obj_descr, obj_descr2])

    descrs = backend.fetch()
    assert len(descrs) == 2
    assert descrs[0].seen == old_seen
    assert descrs[1].seen

    history = backend.fetch_history()
    assert len(history) == 1
    assert history[0].counts == [1, 1, 2]



import logging
import threading
from concurrent.futures import Future
from functools import wraps
from typing import Any
from typing import Callable


logger = logging.getLogger(__name__)


def _wrap_call(func, fut: Future):
    @wraps(func)
    def inner(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
            fut.set_result(result)
            return result
        except Exception as e:
            fut.set_exception(e)
            raise

    return inner


def start_thread(func: Callable, *args: Any, **kwargs: Any) -> Future:
    future: Future = Future()
    should_spawn_thread = True
    try:
        # Gevent greenlets would block all other threads while this one
        # is doing its computation. If we detect gevent has patched
        # threading we run this in a native thread via gevent.threadpool.
        import gevent.monkey

        try:
            if gevent.monkey.is_module_patched("threading"):
                gevent.get_hub().threadpool.spawn(
                    _wrap_call(func, future), *args, **kwargs
                )
                should_spawn_thread = False
        except:
            logger.warning(
                "error using gevent threadpool. Falling back to regular thread",
                exc_info=True,
            )
    except ImportError:
        pass

    if should_spawn_thread:
        thread = threading.Thread(
            name="PyLoot Thread",
            target=_wrap_call(func, future),
            daemon=True,
            args=args,
            kwargs=kwargs,
        )
        thread.start()

    return future



from abc import ABC
from abc import abstractmethod
from typing import Iterable
from typing import List
from typing import Optional

from pyloot.types import ObjectDescriptor
from pyloot.types import ObjectTypeHistory


class BaseBackend(ABC):
    @abstractmethod
    def store(self, object_data: List[ObjectDescriptor]):
        """
        Store object data in the backend
        :param object_data: list of ::class::`ObjectDescriptor`
        """

    @abstractmethod
    def fetch(self, limit: Optional[int]) -> List[ObjectDescriptor]:
        """
        Get all object descriptors

        :param limit: limits number of results returned
        :return: list of ::class::`ObjectDescriptor`
        """

    @abstractmethod
    def fetch_by_id(self, _id: int) -> Optional[ObjectDescriptor]:
        """
        Get object descriptor by memory id

        :param _id: id of object as returned by `id()`
        :return: ::class::`ObjectDescriptor` if found otherwise `None`
        """

    @abstractmethod
    def fetch_children_of(self, _id: int) -> List[ObjectDescriptor]:
        """
        Get child object descriptors of an object given it's memory id

        :param _id: id of parent object as returned by `id()`
        :return: list of ::class::`ObjectDescriptor` if found otherwise an empty list is returned
        """

    @abstractmethod
    def fetch_parents_of(self, _id: int) -> List[ObjectDescriptor]:
        """
        Get child object descriptors of an object given it's memory id

        :param _id: id of parent object as returned by `id()`
        :return: list of ::class::`ObjectDescriptor` if found otherwise an empty list is returned
        """

    @abstractmethod
    def fetch_by_group(
        self, group: str, limit: Optional[int]
    ) -> List[ObjectDescriptor]:
        """
        Get object descriptors by its type name

        :param group: descriptor group
        :param limit: limits number of results returned
        :return: list of ::class::`ObjectDescriptor`
        """

    @abstractmethod
    def fetch_history(self, top: int = 100) -> List[ObjectTypeHistory]:
        """
        Get object history grouped by object_type

        :param top: limit history to the `top` number of entries based on cur count
        :return: list of ::class`ObjectTypeHistory`
        """

    @abstractmethod
    def get_ids(self) -> Iterable[int]:
        """
        Get object ids held by this class which should be ignored when
        collecting objects.

        :return: Iterable of int ids generated by `id()`
        """



from typing import NamedTuple

from pyloot import InMemoryBackend
from pyloot import PyLoot


class Foo(NamedTuple):
    a: int
    b: int


MEM_LK_SIM = []


def test_collection():
    f = Foo(1, 2)
    MEM_LK_SIM.append(f)
    pyloot = PyLoot()
    pyloot.collect_objects()
    backend: InMemoryBackend = pyloot._backend
    assert isinstance(backend, InMemoryBackend)
    objs = backend.fetch()
    assert len(objs) > 0
    obj = backend.fetch_by_id(id(f))
    assert obj
    objs = backend.fetch_by_group("test_collection.Foo")
    assert len(objs) == 1



import sys
from unittest import mock

from pyloot import cli


@mock.patch("pyloot.cli.PyLootServer.serve_forever")
def test_base_cli(mocked_serve):
    with mock.patch.object(sys, "argv", ["pyloot"]):
        cli.main()
    mocked_serve.assert_called_once_with(host="0.0.0.0", port=8000)


@mock.patch("pyloot.cli.PyLootServer.serve_forever")
def test_cli_override_port(mocked_serve):
    with mock.patch.object(sys, "argv", ["pyloot", "-p", "8081"]):
        cli.main()
    mocked_serve.assert_called_once_with(host="0.0.0.0", port=8081)


@mock.patch("pyloot.cli.PyLootServer.serve_forever")
def test_cli_override_port_host(mocked_serve):
    with mock.patch.object(sys, "argv", ["pyloot", "-p", "8081", "-h", "127.0.0.1"]):
        cli.main()
    mocked_serve.assert_called_once_with(host="127.0.0.1", port=8081)



from setuptools import setup


setup(
    name="pyloot",
    versioning="distance",
    setup_requires="setupmeta",
    author="Michael Chase",
    description="Multiprocessing compatible memory leak debugger inspired by dozer/dowser",
    keywords="memory profiler multiprocessing wsgi asgi",
    url="https://github.com/reallistic/pyloot",
    install_requires=["WebOb>=1.2", "importlib-resources"],
    extras_require={"test": ["pytest"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: Public Domain",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.6",
    entry_points=dict(console_scripts=["pyloot=pyloot.cli:main"]),
)
